/*
 * Copyright (c) 2001 Zajcev Evgeny <lg@rrf.ru>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * main.c [v1.0]
 */

#include <gtk/gtk.h>

#include <sys/time.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "virt.h"
#include "stats.h"
#include "fonts.h"
#include "text.h"
#include "time.h"
#include "iface.h"

#define L_LINES (L_HEAD_LEN+1+MAX_STRS_IN_EX*4)
extern int      exercises;
/*
 * global virt variables 
 */
struct timeval  speed_bt,
                speed_et,
                difft,
                ft,
                st,
                dr_average;
char           *virt_level_lines[L_LINES];
GtkWidget      *window;
struct virt_pstat virt_speed;	/* it's bad */
struct virt_pstat virt_l_errs;
struct virt_pstat virt_l_dr;
struct virt_pstat virt_errs;
struct virt_pstat virt_dr;

struct virt_slist *vslh,
               *vsl;		/* head and current ex virt_slists */
struct virt_user_stats vus;
struct virt_sl  virt_str;

struct key_top_list virt_ktl;

int             virt_use_beep = 0;
int		enable_beep_on_error=0;
int             enable_ktop = 0;
int             enable_ktop_nk = 0;
int             beep_tag;	/* for beep */
char            virt_uline[STR_LEN];

int             current_level;
int             current_ex;
GtkWidget      *virt_ml,
               *virt_hdr,
               *key_top;

GtkItemFactory *fac;

void
close_virt(GtkWidget * widget, gpointer data)
{
    virt_wuc(&vus);
    gtk_main_quit();
}

void
virt_save(void)
{
    printf("saved\n");
    virt_wuc(&vus);
}

/*
 * used in "/Options/Show Keytop"
 * enable_ktop = -1 used only for first check
 */
void
hide_show_keytop(void)
{
    switch (enable_ktop) {
    case -1:{
	    enable_ktop = 1;
	    gtk_widget_set_usize(window, MW_W, MW_H + KT_H);
	    break;
	}
    case 1:{
	    enable_ktop = 0;
	    gtk_widget_hide(key_top);
	    gtk_widget_set_usize(window, MW_W, MW_H);
	    break;
	}
    default:{
	    enable_ktop = 1;
	    gtk_widget_show(key_top);
	    gtk_widget_set_usize(window, MW_W, MW_H + KT_H);
	}
    }
}

void
virt_show_nk(void)
{
    if (enable_ktop_nk == 1)
	virt_ktop_key(virt_str.str[virt_str.nexc], 1);
}

void
hide_show_next_key(void)
{
  switch (enable_ktop_nk) {
  case -1:
    enable_ktop_nk = 1;
    break;
  case 0: enable_ktop_nk = 1;
    virt_show_nk();
    break;
  default: virt_ktop_key(virt_str.str[virt_str.nexc], 0);
    enable_ktop_nk = 0;
  }
}

GtkItemFactoryEntry menu_items[] = {
    {"/Game/Save", "<control>S", virt_save, 1, "<Item>"} ,
    {"/Game/", NULL, NULL, 0, "<Separator>"} ,
    {"/Game/Quit", "<control>Q", close_virt, 0, "<Item>"} ,
    {"/Options/Show Keytop", NULL, hide_show_keytop, 0, "<CheckItem>"} ,
    {"/Options/Show Next Key", NULL, hide_show_next_key, 0, "<CheckItem>"}
};

void
main_menu(GtkWidget * win, GtkWidget ** menubar)
{
    int             nm_i = sizeof(menu_items) / sizeof(menu_items[0]);

    fac = gtk_item_factory_new(GTK_TYPE_MENU_BAR, "<Main>", NULL);
    gtk_item_factory_create_items(fac, nm_i, menu_items, NULL);
    gtk_window_add_accel_group(GTK_WINDOW(window), fac->accel_group);
    if (enable_ktop == 1) {
	enable_ktop = -1;
	gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM
				       (gtk_item_factory_get_item
					(fac, "/Options/Show Keytop")),
				       TRUE);
    }
    if (enable_ktop_nk == 1) {
	enable_ktop_nk = -1;
	gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM
				       (gtk_item_factory_get_item
					(fac, "/Options/Show Next Key")),
				       TRUE);
    }

    if (menubar)
	*menubar = fac->widget;

}

void
virt_fill_vs(void)
{
    virt_str.nexc = 0;
    if (current_level < vsl->q_strs + 1)
	strcpy(virt_str.str, vsl->strings[current_level]);
    else {
	fprintf(stderr, "level %d is too big, using 0\n", current_level);
	current_level = 0;
	strcpy(virt_str.str, vsl->strings[current_level]);
    }
}

void
virt_init_vsl(void)
{
    int             i;

    vsl = vslh;
    for (i = 1; i < current_ex; i++)
	vsl = vsl->next;
}

void
refresh_stats(void)
{
    vus.l_errs = vus.l_dr = -1;
    vus.errs--;
    vus.dr--;
    virt_update_errs(&vus);
    virt_update_dr(&vus);
    virt_update_speed(&vus);
    virt_update_level_lines(vsl->q_strs, current_level);
}

void
virt_next_ex(void)
{
    vus.l_errs = vus.errs = vus.l_dr = vus.dr = vus.speed = 0;
    vsl = vsl->next;
    current_level = 0;
    vus.ex = current_ex += 1;
    virt_fill_vs();
    gtk_label_set_text((gpointer) virt_hdr, vsl->header);
    virt_show_nk();
}

/*
 * update virt main label with filling virt_str.nexc by 0x20 
 */
void
change_virt_ml(GtkWidget * widget, gpointer data)
{
    int             next_cn;

    next_cn = virt_str.nexc;
    /*
     * enable dr counter 
     */
    if (next_cn == 0) {
	gettimeofday(&ft, (struct timezone *) NULL);
	speed_bt = ft;
    }
    if (next_cn == 1) {
	gettimeofday(&st, (struct timezone *) NULL);
	dr_average = diff_time(ft, st);
    }
    if (next_cn > 1) {
	gettimeofday(&ft, (struct timezone *) NULL);
	difft = diff_time(st, ft);
	dr_average = add_time(dr_average, difft);
	dr_average = div_time(dr_average, 2);
	st = ft;
	/*
	 * check disrhithm 
	 */
	if (compare_time(div_time(diff_time(difft, dr_average), DR_COEF),
			 dr_average) == 1)
	    virt_update_dr(&vus);
	if ((next_cn % 7) == 0) {
	    speed_et = ft;
	    if (vus.speed == 0)
		vus.speed =
		    60.0 * next_cn / (speed_et.tv_sec - speed_bt.tv_sec);
	    else
		vus.speed =
		    (60.0 * next_cn / (speed_et.tv_sec - speed_bt.tv_sec)
		     + vus.speed) / 2;
	    virt_update_speed(&vus);
	}
    }
    if ((next_cn == 0) && (virt_use_beep == 1)) {
	gdk_beep();
	beep_tag =
	    gtk_timeout_add(60 * 1000 / 200, (GtkFunction) gdk_beep, NULL);
    }
    if (virt_str.str[next_cn + 1] != '\0') {
	virt_str.str[next_cn] = 0x20;
	virt_str.nexc += 1;
	virt_show_nk();
    } else {
	if (virt_use_beep == 1)
	    gtk_timeout_remove(beep_tag);
	/*
	 * check results 
	 */
	if ((virt_check_stats(1, &vus) >= 0)
	    && (virt_check_stats(2, &vus) >= 0)) {
	    current_level += 1;
	    if (current_level == vsl->q_strs) {
		if ((virt_check_stats(3, &vus) >= 0) &&
		    (virt_check_stats(4, &vus) >= 0) &&
		    (virt_check_stats(5, &vus) >= 0))
		    virt_next_ex();
		else {
		    current_level = 0;
		    vus.speed = vus.l_errs = vus.errs = vus.l_dr = vus.dr =
			vus.speed = 0;
		}
	    }
	}
	virt_fill_vs();
	refresh_stats();
	virt_show_nk();
	if (virt_use_beep == 1)
	    gtk_timeout_remove(beep_tag);
    }
    gtk_label_set_text((gpointer) virt_ml, (virt_str.str));
    /*
     * set uline 
     */
    memset(virt_uline, 0x20, STR_LEN);
    virt_uline[virt_str.nexc] = '_';
    gtk_label_set_pattern(GTK_LABEL(virt_ml), virt_uline);
}

void
virt_key_pressed(GtkWidget * wid, GdkEventKey * event)
{
    if (event->keyval > SP_BEGIN)
	return;
    if ((event->keyval == virt_str.str[virt_str.nexc]) ||
	(event->keyval == 256 + virt_str.str[virt_str.nexc])) {
	if (enable_ktop_nk == 1)
	    virt_ktop_key(event->keyval, 0);
	change_virt_ml(virt_ml, NULL);
    } else {
	if (enable_beep_on_error == 1)
	  gdk_beep();
	virt_update_errs(&vus);
    }
}

void
usage(char *name)
{
    fprintf(stderr, "usage: %s [-nkb] [-f <fontname>]\n", name);
}

/*
 * main routine 1) create widgets, 2) connect events 
 */
int
main(int argc, char *argv[])
{
    int             ch;
    char            font_n[200];
    char           filen[200];
    GtkWidget      *stat_box,
                   *all_box,
                   *hbox,
                   *vbox;
    GtkWidget      *separ,
                   *ll,
                   *menubar;
    GtkWidget      *speed,
                   *l_dr,
                   *dr,
                   *l_errs,
                   *errs;
    int             i;

    strncpy(font_n, DS_FONT, 200);
    strncpy(filen, PREFIX "/share/virt/virt.eng", 200);
    while ((ch = getopt(argc, argv, "rbknf:")) != -1)
	switch (ch) {
	case 'f':     /* font in main label */
	    strncpy(font_n, optarg, 200);
	    break;
	case 'k':     /* show keytop */
	    enable_ktop = 1;
	    break;
	case 'n':     /* show next keytop key */
	    enable_ktop_nk = 1;
	    break;
	case 'b':
	    enable_beep_on_error = 1;
	    break;
	case 'r':
	    strncpy(filen, PREFIX "/share/virt/virt.rus", 200);
	    break;
	case '?':
	default:
	    usage(argv[0]);
	    exit(0);
	}

    /* use rc file for keytop buttons and for future */
    gtk_rc_add_default_file(PREFIX "/share/virt/virtrc");
    gtk_init(&argc, &argv);
    gdk_rgb_init();

    if ((vslh = virt_open_file(filen)) == NULL) {
	gtk_main();
	return 1;
    }
    /* some init information [bad] */
    vsl = vslh;
    current_level = 0;
    virt_ruc(&vus);
    current_ex = vus.ex;
    virt_init_vsl();
    virt_fill_vs();

    window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_widget_set_usize(window, MW_W, MW_H);
    /*
     * set policy for immediatly resizing
     * window when usize is changed
     */
    gtk_window_set_policy(GTK_WINDOW(window), FALSE, TRUE, TRUE);
    gtk_window_set_title(GTK_WINDOW(window), "Virtuoso");
    /* connect some events */
    gtk_signal_connect(GTK_OBJECT(window), "destroy",
		       GTK_SIGNAL_FUNC(close_virt), NULL);

    gtk_signal_connect(GTK_OBJECT(window), "delete_event",
		       GTK_SIGNAL_FUNC(close_virt), NULL);

    gtk_signal_connect(GTK_OBJECT(window), "key_press_event",
		       GTK_SIGNAL_FUNC(virt_key_pressed), NULL);
    gtk_widget_set_events(window, GDK_KEY_PRESS);
    /***/
    gtk_container_set_border_width(GTK_CONTAINER(window), 1);
    gtk_widget_realize(window);

    all_box = gtk_vbox_new(FALSE, 0);
    gtk_container_add(GTK_CONTAINER(window), all_box);

    /* create menu bar */
    main_menu(window, &menubar);
    gtk_box_pack_start(GTK_BOX(all_box), menubar, FALSE, FALSE, 0);
    gtk_widget_show(menubar);

    /* display header of exercise */
    virt_hdr = gtk_label_new(vsl->header);
    gtk_widget_set_name(virt_hdr, "virt_hdr");
    gtk_label_set_justify(GTK_LABEL(virt_hdr), GTK_JUSTIFY_CENTER);
    gtk_widget_set_usize(virt_hdr, MW_W, HD_H);
    gtk_box_pack_start(GTK_BOX(all_box), virt_hdr, FALSE, FALSE, 4);
    gtk_widget_show(virt_hdr);

    /* create main label with specifyed font */
    virt_ml = gtk_label_new(virt_str.str);
    gtk_widget_set_name(virt_ml, "virt_ml");
    gtk_label_set_pattern(GTK_LABEL(virt_ml), "_");
    gtk_widget_set_usize(virt_ml, MW_W, ML_H);
    gtk_label_set_justify(GTK_LABEL(virt_ml), GTK_JUSTIFY_CENTER);
    virt_set_font(virt_ml, font_n);
    gtk_box_pack_start(GTK_BOX(all_box), virt_ml, FALSE, FALSE, 0);
    gtk_widget_show(virt_ml);


    separ = gtk_hseparator_new();
    gtk_box_pack_start(GTK_BOX(all_box), separ, FALSE, FALSE, 0);
    gtk_widget_show(separ);

    stat_box = gtk_hbox_new(TRUE, 0);

    /* virt level lines */
    vbox = gtk_vbox_new(FALSE, 0);
    for (i = 0; i < L_LINES; i++) {
	virt_level_lines[i] = (char *) malloc(200);
    }
    ll = virt_show_level_lines(vsl->q_strs, current_level);
    gtk_box_pack_start(GTK_BOX(vbox), ll, FALSE, FALSE, STAT_S);
    gtk_widget_show(ll);
    gtk_box_pack_start(GTK_BOX(stat_box), vbox, TRUE, TRUE, 0);
    gtk_widget_show(vbox);

    /* speed stat */
    vbox = gtk_vbox_new(FALSE, 0);
    speed = virt_show_speed(vus);
    gtk_box_pack_start(GTK_BOX(vbox), speed, FALSE, FALSE, STAT_S);
    gtk_widget_show(speed);
    gtk_box_pack_start(GTK_BOX(stat_box), vbox, TRUE, TRUE, 0);
    gtk_widget_show(vbox);

    /* level stats */
    vbox = gtk_vbox_new(FALSE, 0);
    l_dr = virt_show_level_dr(vus);
    gtk_box_pack_start(GTK_BOX(vbox), l_dr, FALSE, FALSE, STAT_S);
    gtk_widget_show(l_dr);

    l_errs = virt_show_level_errs(vus);
    gtk_box_pack_start(GTK_BOX(vbox), l_errs, FALSE, FALSE, STAT_S);
    gtk_widget_show(l_errs);

    gtk_box_pack_start(GTK_BOX(stat_box), vbox, TRUE, TRUE, 0);
    gtk_widget_show(vbox);

    /* ex stats */
    vbox = gtk_vbox_new(FALSE, 0);
    dr = virt_show_ex_dr(vus);
    gtk_box_pack_start(GTK_BOX(vbox), dr, FALSE, FALSE, STAT_S);
    gtk_widget_show(dr);

    errs = virt_show_ex_errs(vus);
    gtk_box_pack_start(GTK_BOX(vbox), errs, FALSE, FALSE, STAT_S);
    gtk_widget_show(errs);

    gtk_box_pack_start(GTK_BOX(stat_box), vbox, TRUE, TRUE, 0);
    gtk_widget_show(vbox);

    /* all stats box is ok - show it */
    gtk_box_pack_start(GTK_BOX(all_box), stat_box, TRUE, TRUE, 0);
    gtk_widget_show(stat_box);


    /* create keytop */
    hbox = gtk_hbox_new(TRUE, 0);
    key_top = virt_keytop(&virt_ktl);
    gtk_box_pack_start(GTK_BOX(hbox), key_top, FALSE, FALSE, 0);
    gtk_box_pack_start(GTK_BOX(all_box), hbox, FALSE, FALSE, 0);
    gtk_widget_show(hbox);
    if (enable_ktop == 1)
	gtk_widget_show(key_top);

    virt_show_nk();
    gtk_widget_show(all_box);
    gtk_widget_show(window);

    gtk_main();
    return (0);
}
