#include <gtk/gtk.h>
#include <string.h>
#include <stdio.h>
#include "virt.h"
#include "numbers.h"
#include "stats.h"

GtkWidget      *ll_pix;		/* Widget that shows virt_level_lines */

static char    *l_head[L_HEAD_LEN] = {
    " 3 1",
    "` c None",
    "d c #008000",
    "n c #FF0000"
};
/*
 * 99 len 
 */
static char     l_void[] =
    "```````````````````````````````````````````````````````````````````````````````````````````````````";
static char     l_line[] =
    "nnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnn";
static char     l_dline[] =
    "ddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddd";

/*
 * statistic table *
 * +------------------+------------+------------+------------+ 
 * | skill            |     2      |       1    |     0      | 
 * +------------------+------------+------------+------------+ 
 * |exersise  speed   | >= 200 s/m | >= 150 s/m | >= 100 s/m |
 * |exersise disrhythm| * <= 20 %  | <= 30 %    | <= 40 %    |
 * |exersise errors   | <= 0.5 %   | <= 1 %     | <= 2 %     |
 * |level speed       | >= 0 s/m   | >= 0 s/m   | >= 0 s/m   |
 * |level disrhythm   | <= 20 %    | <= 30 %    | <= 40 %    |
 * |level errors      | <= 0.5 %   | <= 2 %     | <= 4 %     |
 * +------------------+------------+------------+------------+ 
 */
float           glov_stats[3][6] = {
    {0, 0.4, 0.04, 100, 0.4, 0.02},	/* skill 0 */
    {0, 0.3, 0.02, 150, 0.3, 0.01},	/* skill 1 */
    {0, 0.2, 0.005, 200, 0.2, 0.005}	/* skill 2 */
};

GtkWidget      *
virt_show_level_errs(struct virt_user_stats vus)
{
    GtkWidget      *label,
                   *hbox;

    hbox = gtk_hbox_new(FALSE, 0);
    label = gtk_label_new("Level errors: ");
    gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
    gtk_widget_show(label);
    virt_l_errs = print_num(vus.l_errs, DEF_OK_COL);
    gtk_box_pack_start(GTK_BOX(hbox), virt_l_errs.wid, FALSE, FALSE, 0);
    gtk_widget_show(virt_l_errs.wid);

    return (hbox);
}

GtkWidget      *
virt_show_level_dr(struct virt_user_stats vus)
{
    GtkWidget      *label,
                   *hbox;

    hbox = gtk_hbox_new(FALSE, 0);
    label = gtk_label_new("Level disrhithm: ");
    gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
    gtk_widget_show(label);
    virt_l_dr = print_num(vus.l_dr, DEF_OK_COL);
    gtk_box_pack_start(GTK_BOX(hbox), virt_l_dr.wid, FALSE, FALSE, 0);
    gtk_widget_show(virt_l_dr.wid);

    return (hbox);
}

GtkWidget      *
virt_show_ex_errs(struct virt_user_stats vus)
{
    GtkWidget      *label,
                   *hbox;

    hbox = gtk_hbox_new(FALSE, 0);
    label = gtk_label_new("Exercise errors: ");
    gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
    gtk_widget_show(label);
    virt_errs = print_num(vus.errs, DEF_OK_COL);
    gtk_box_pack_start(GTK_BOX(hbox), virt_errs.wid, FALSE, FALSE, 0);
    gtk_widget_show(virt_errs.wid);

    return (hbox);
}

GtkWidget      *
virt_show_ex_dr(struct virt_user_stats vus)
{
    GtkWidget      *label,
                   *hbox;

    hbox = gtk_hbox_new(FALSE, 0);
    label = gtk_label_new("Exercise disrhithm: ");
    gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
    gtk_widget_show(label);
    virt_dr = print_num(vus.dr, DEF_OK_COL);
    gtk_box_pack_start(GTK_BOX(hbox), virt_dr.wid, FALSE, FALSE, 0);
    gtk_widget_show(virt_dr.wid);

    return (hbox);
}

GtkWidget      *
virt_show_speed(struct virt_user_stats vus)
{
    GtkWidget      *hbox,
                   *label;

    hbox = gtk_hbox_new(FALSE, 0);
    label = gtk_label_new("Speed: ");
    gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
    gtk_widget_show(label);
    virt_speed = print_num(vus.speed, DEF_OK_COL);
    gtk_box_pack_start(GTK_BOX(hbox), virt_speed.wid, FALSE, FALSE, 0);
    gtk_widget_show(virt_speed.wid);

    return (hbox);
}

int
itoa(int num, char *str)
{
    int             dig,
                    i = 0,
                    ch,
                    ret;

    do {
	dig = num % 10;
	str[i] = dig + 48;	/* digit to char */
	num /= 10;
	i++;
    } while (num > 0);
    str[i--] = '\0';
    ret = i;
    /*
     * reverse string 
     */
    for (dig = 0; dig < i; dig++, i--) {
	ch = str[dig];
	str[dig] = str[i];
	str[i] = ch;
    }
    return i;
}

void
virt_gen_level_xpm(int lines, int done_lines)
{
    int             i;
    char           *p;

    /*
     * fill head 
     */
    strcpy(virt_level_lines[0], "99 ");
    p = virt_level_lines[0] + 3;
    itoa(lines * 4, p);
    strcat(virt_level_lines[0], l_head[0]);
    for (i = 1; i < L_HEAD_LEN; i++)
	strcpy(virt_level_lines[i], l_head[i]);

    /*
     * fill not done lines 
     */
    for (; i < (L_HEAD_LEN + 4 * (lines - done_lines)); i++) {
	strcpy(virt_level_lines[i++], l_line);
	strcpy(virt_level_lines[i++], l_line);
	strcpy(virt_level_lines[i++], l_void);
	strcpy(virt_level_lines[i], l_void);
    }

    for (; i < (L_HEAD_LEN + 4 * lines); i++) {
	strcpy(virt_level_lines[i++], l_dline);
	strcpy(virt_level_lines[i++], l_dline);
	strcpy(virt_level_lines[i++], l_void);
	strcpy(virt_level_lines[i], l_void);
    }

}

GtkWidget      *
virt_show_level_lines(int l, int dl)
{
    GtkWidget      *vbox;
    GtkStyle       *style;
    GdkPixmap      *r_pix;
    GdkBitmap      *mask;

    style = gtk_widget_get_style(window);
    virt_gen_level_xpm(l, dl);

    vbox = gtk_vbox_new(FALSE, 0);
    r_pix = gdk_pixmap_create_from_xpm_d(window->window, &mask,
					 &style->bg[GTK_STATE_NORMAL],
					 (gchar **) virt_level_lines);
    ll_pix = gtk_pixmap_new(r_pix, mask);
    gtk_box_pack_start(GTK_BOX(vbox), ll_pix, FALSE, FALSE, 0);
    gtk_widget_show(ll_pix);

    return vbox;
}

void
virt_update_level_lines(int l, int dl)
{
    GtkStyle       *style;
    GdkPixmap      *r_pix;
    GdkBitmap      *mask;

    style = gtk_widget_get_style(window);
    virt_gen_level_xpm(l, dl);

    r_pix = gdk_pixmap_create_from_xpm_d(window->window, &mask,
					 &style->bg[GTK_STATE_NORMAL],
					 (gchar **) virt_level_lines);
    gtk_pixmap_set(GTK_PIXMAP(ll_pix), r_pix, mask);
}

float
virt_check_stats(int what, struct virt_user_stats *vus)
{
    float           snum;

    switch (what) {
    case 1:{
	    snum = vus->l_errs * 100.0 / strlen(virt_str.str);
	    if (snum > 100 * glov_stats[vus->skill][2])
		return -snum;
	    return snum;
	}
    case 2:{
	    snum = vus->l_dr * 100.0 / strlen(virt_str.str);
	    if (snum > 100 * glov_stats[vus->skill][1])
		return -snum;
	    return snum;
	}
    case 3:{
	    snum = vus->errs * 100.0 / vsl->chars;
	    if (snum > 100 * glov_stats[vus->skill][5])
		return -snum;
	    return snum;
	}
    case 4:{
	    snum = vus->dr * 100.0 / vsl->chars;
	    if (snum > 100 * glov_stats[vus->skill][4])
		return -snum;
	    return snum;
	}
    case 5:{
	    if (vus->speed < glov_stats[vus->skill][3])
		return -vus->speed;
	    return vus->speed;
	}
    }
    return -1.0;
}

void
virt_update_errs(struct virt_user_stats *vus)
{
    float           er;

    vus->l_errs++;
    vus->errs++;

    er = virt_check_stats(1, vus);
    if (er < 0)
	update_num(&virt_l_errs, -er, DEF_FA_COL);
    else
	update_num(&virt_l_errs, er, DEF_OK_COL);

    er = virt_check_stats(3, vus);
    if (er < 0)
	update_num(&virt_errs, -er, DEF_FA_COL);
    else
	update_num(&virt_errs, er, DEF_OK_COL);
}

void
virt_update_dr(struct virt_user_stats *vus)
{
    float           dr;

    vus->l_dr++;
    vus->dr++;

    dr = virt_check_stats(2, vus);
    if (dr < 0)
	update_num(&virt_l_dr, -dr, DEF_FA_COL);
    else
	update_num(&virt_l_dr, dr, DEF_OK_COL);

    dr = virt_check_stats(4, vus);
    if (dr < 0)
	update_num(&virt_dr, -dr, DEF_FA_COL);
    else
	update_num(&virt_dr, dr, DEF_OK_COL);
}

void
virt_update_speed(struct virt_user_stats *vus)
{
    float           speed;

    speed = virt_check_stats(5, vus);
    if (speed < 0)
	update_num(&virt_speed, -speed, DEF_FA_COL);
    else
	update_num(&virt_speed, speed, DEF_OK_COL);
}
